pro destripe, drifts_all, flag_valid_scans, map_back=map_back, smallfield=smallfield, $
              size_deg=size_deg, file_scalars=file_scalars, ind_sc_coarse=ind_sc_coarse, $
              ind_scans=ind_scans, ind_subscans=ind_subscans, maxnoise=maxnoise, $
              cut_start=cut_start, galactic=galactic, time_1=time_1, t_proc=t_proc


;;; computes residual linear baselines from the redundancy

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2012 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

crs = string(13b) + string(10b)


if ((size_deg lt 1. and nscans le 20) or nt lt 50000L) then ns = 49 else ns = 0
size_lim_ref = 0.75

cube_signal = fltarr(dimx_iter, dimy_iter, nb, nscans)
cube_weight = fltarr(dimx_iter, dimy_iter, nb, nscans)
cube_subscan = lonarr(dimx_iter, dimy_iter, nb, nscans)
dimxy = dimx_iter * dimy_iter
dimxynb = dimxy * nb
tab_nb = dimxy * lindgen(nb)

tab_scan = lonarr(ind_scans(1))
for i = 1, nscans - 1 do tab_scan = [temporary(tab_scan), i + lonarr(ind_scans(i+1) - ind_scans(i))]

tab_subscan = lonarr(ind_subscans(1))
for i = 1L, nsubscans - 1L do tab_subscan = [temporary(tab_subscan), i + lonarr(ind_subscans(i+1) - ind_subscans(i))]
tab_subscan += 1L

restore, file_times
w = where(flag_nza lt 1, cw)
if cw ge 1 then tab_subscan(w) = 0L
for i = 0L, nsubscans - 1L do begin
   w = where(flag_nza(ind_subscans(i):ind_subscans(i+1)-1) lt 1, cw)
   if cw ge 1 then tab_subscan(ind_subscans(i)+max([0L, w(0)-200L]):min([ind_subscans(i+1)-1, ind_subscans(i)+w(cw-1)+200L])) = i + 1L
endfor

weight = 1. / maxnoise^2.


;;; remove duplicate offset grid:
prepare_proj_coarse, file_scalars=file_scalars, /onegrid, ind_subscans=ind_subscans, ind_scans=ind_scans



restore, file_signal

restore, file_projcoarse

restore, file_vscan

restore, file_time_coarse


for k = 0L, nproj - 1 do begin
   revind_0 = revind(wproj(k))
   revind_1 = revind(wproj(k)+1)
   cp = revind_1 - revind_0
   wp = revind(revind_0:revind_1-1)
   wb = wp / nt
   ws = tab_scan(wp mod nt)
   wbs = wb + nb * ws
   ind_bs = sort(wbs)
   wp = wp(ind_bs)
   wb = wb(ind_bs)
   ws = ws(ind_bs)
   wbs = wbs(ind_bs)
   wt = wp mod nt
   ind_bs = uniq(wbs)
   cbs = n_elements(ind_bs)

   weight_sample = weight(tab_scan(wp mod nt) + nscans * (wp / nt))

   tab_ind = [-1, ind_bs]
   tab_b = wb(ind_bs)
   tab_s = ws(ind_bs)
   for ibs = 0L, cbs - 1 do begin
      wplane = tab_ind(ibs) + 1 + lindgen(tab_ind(ibs+1) - tab_ind(ibs))
      vectpos_plane = wproj(k) + dimxy * tab_b(ibs) + dimxynb * tab_s(ibs)
      cube_weight(vectpos_plane) += total(weight_sample(wplane))
      cube_signal(vectpos_plane) += total(signal(wp(wplane)) * weight_sample(wplane))
      cube_subscan(vectpos_plane) = median(tab_subscan(wt(wplane)))
   endfor
endfor

wcov = where(cube_weight gt 0.)
cube_signal(wcov) = cube_signal(wcov) / cube_weight(wcov)
w = where(finite(weight) gt 0)
cube_weight = cube_weight / mean(weight(w))
w = where(cube_weight le 0.)
cube_signal(w) = !values.f_nan

prepare_proj_coarse, file_scalars=file_scalars



;;; determine whether the field is covered by largely disjoint scans or not:

weight_scans = total(cube_weight, 3, /nan)
w = where(weight_scans gt 0.)
weight_scans /= median(weight_scans(w))
weight_cross = 1
tab_cw = lonarr(nscans)
for i = 0, nscans - 1 do begin
   weight_cross *= weight_scans(*, *, i)
   w = where(weight_scans(*, *, i) gt 0., cw)
   tab_cw(i) = cw
endfor
w = where(weight_cross gt 0., cw)
if cw lt min(tab_cw) / 4. then disjoint = 1
weight_scans = 0
weight_cross = 0
w = 0



;;; exclude sources using the same mask as baselines.pro:

if keyword_set(smallfield) then begin
   restore, file_maps
   w = where(map_weight gt 0., cw, complement=wz)
   ind_sort = reverse(sort(map_signal(w)))
   map_back = map_signal
   map_back(wz) = !values.f_nan
   map_back(w(ind_sort(0:cw/20))) = !values.f_nan
endif
map_mask = fltarr(dimx_iter, dimy_iter)
nx = (dimx / fact) * fact
ny = (dimy / fact) * fact
map_mask(0:dimx/fact-1, 0:dimy/fact-1) = rebin(map_back(0:nx-1, 0:ny-1), dimx/fact, dimy/fact)
w = where(finite(map_mask) lt 1 and total(cube_weight(*, *, *, 0), 3) gt 0. $
   and total(cube_weight(*, *, *, 1), 3) gt 0., cw)
if cw ge 1 then begin
   wz_init = reform(rebin(w, cw, nb) + (tab_nb ## (1L + lonarr(cw))), cw*nb)
   for s = 0L, nscans - 1 do begin
      wz = wz_init + dimxynb * s
      cube_weight(wz) = 0.
      cube_subscan(wz) = 0.
   endfor
endif



;;; create maps for single scans,
;;; and build reference maps for even and odd scans, excluding potentially deviant scans:

cube_weight_ref = cube_weight
if (nscans ge 4 and keyword_set(cut_start)) then begin
   wcut = where(cut_start gt 0, cw_cut)
   for c = 0, cw_cut - 1 do begin
      w = where(ind_subscans ge ind_scans(wcut(c)))
      subs = w(0)
      w = where(cube_subscan eq subs + 1, cw)
      if cw ge 1 then cube_weight_ref(w) = 0.
   endfor
endif

map_scans = total(cube_signal * cube_weight_ref, 3, /nan)
weight_scans = total(cube_weight_ref, 3)
wdef_scans = where(weight_scans gt 0., complement=wz_scans)
map_scans(wdef_scans) /= weight_scans(wdef_scans)
map_scans(wz_scans) = !values.f_nan

flag_valid_scans = 1 + intarr(nscans)
if (not keyword_set(disjoint)) then if nscans gt 4 then repeat begin
   nexc = 0
   ind_valid = where(flag_valid_scans gt 0 and (indgen(nscans) mod 2) eq 0, c_valid)
   map_median = median(map_scans(*, *, ind_valid), dimension=3, /even)
   dev = sqrt(median((map_scans(*, *, ind_valid) - rebin(map_median, dimx_iter, dimy_iter, c_valid))^2.))
   if c_valid gt 10 then dev *= sqrt(float(c_valid) / 10.)
   for s = 0, nscans - 1, 2 do if flag_valid_scans(s) gt 0 then begin
      dev_scan = median(abs(map_scans(*, *, s) - map_median))
      if dev_scan gt 4. * dev then begin
         print, '     excluding deviant scan ' + chain(s) + ' from reference maps: ndev = ' + chain(dev_scan / dev, 1)
         flag_valid_scans(s) = 0
         nexc = 1
      endif
   endif
   ind_valid = where(flag_valid_scans gt 0 and (indgen(nscans) mod 2) eq 1, c_valid)
   map_median = median(map_scans(*, *, ind_valid), dimension=3, /even)
   dev = sqrt(median((map_scans(*, *, ind_valid) - rebin(map_median, dimx_iter, dimy_iter, c_valid))^2.))
   if c_valid gt 10 then dev *= sqrt(float(c_valid) / 10.)
   for s = 1, nscans - 1, 2 do if flag_valid_scans(s) gt 0 then begin
      dev_scan = median(abs(map_scans(*, *, s) - map_median))
      if dev_scan gt 4. * dev then begin
         print, '     excluding deviant scan ' + chain(s) + ' from reference maps: ndev = ' + chain(dev_scan / dev, 1)
         flag_valid_scans(s) = 0
         nexc = 1
      endif
   endif
endrep until nexc eq 0
ind_even = where((indgen(nscans) mod 2) eq 0 and flag_valid_scans gt 0, n_even)
ind_odd = where((indgen(nscans) mod 2) eq 1 and flag_valid_scans gt 0., n_odd)
ind_valid = where(flag_valid_scans gt 0)


if n_even ge 2 then begin
   ima_even = total(map_scans(*, *, ind_even) * weight_scans(*, *, ind_even), 3, /nan)
   weight_even = total(weight_scans(*, *, ind_even), 3)
   w_even = where(weight_even gt 0.)
   ima_even(w_even) /= weight_even(w_even)
endif else begin
   ima_even = map_scans(*, *, ind_even)
   weight_even = weight_scans(*, *, ind_even)
   w_even = where(weight_even gt 0.)
endelse
if n_odd ge 2 then begin
   ima_odd = total(map_scans(*, *, ind_odd) * weight_scans(*, *, ind_odd), 3, /nan)
   weight_odd = total(weight_scans(*, *, ind_odd), 3)
   w_odd = where(weight_odd gt 0.)
   ima_odd(w_odd) /= weight_odd(w_odd)
endif else begin
   ima_odd = map_scans(*, *, ind_odd)
   weight_odd = weight_scans(*, *, ind_odd)
   w_odd = where(weight_odd gt 0.)
endelse

weight_ref = total(weight_scans(*, *, ind_valid), 3)
if nscans lt 4 then begin
   wz = where(weight_even * weight_odd le 0.)
   weight_ref(wz) = 0.
endif
wdef_ref = where(weight_ref gt 0., complement=wz_ref)



restore, file_pixcoords
indxy_coarse = floor(x_bolos / fact) + dimx_iter * floor(y_bolos / fact)
x_bolos = 0
y_bolos = 0

if inst eq 'spire' then begin
   min_nloop = 4
   max_nloop = 8
   nloop_switch = 2
   noise_level = median(maxnoise)
endif else begin
   min_nloop = 2
   if (keyword_set(smallfield) and size_deg lt 1.) then max_nloop = 4 else max_nloop = 3
   nloop_switch = 2
   noise_level = 3. * median(maxnoise)
endelse

ima_diff = fltarr(dimx_iter, dimy_iter)

simul = fltarr(nt, nb)



;;; first rectify each scan:

nloop = 4
if keyword_set(galactic) then for loop = 0, nloop - 1 do begin
   for s = 0, nscans - 1 do begin

      if loop lt nloop / 2 then begin
         if (s mod 2) eq 0 then map_ref = ima_odd else map_ref = ima_even
      endif

      indxy_scan = indxy_coarse(ind_scans(s):ind_scans(s+1)-1, *)
      diff_scan = map_scans(dimxy*s+indxy_scan) - map_ref(indxy_scan)
      wb = where(finite(maxnoise(s, *)) gt 0)
      diff = total(diff_scan(*, wb), 2, /nan)
      norm = total(finite(diff_scan(*, wb)), 2)
      w = where(norm gt 0.)
      diff(w) /= norm(w)
      time_ref = time_s(ind_scans(s)+w(0))
      fit_scan = absdev_fit(float(-time_ref + time_s(ind_scans(s)+w)), diff(w), coef=coef, $
         noise_level=noise_level, /sym, /dyn, ns=ns)
      simul_scan = coef(0) + coef(1) * float(-time_ref + time_s(ind_scans(s):ind_scans(s+1)-1))
      simul(ind_scans(s):ind_scans(s+1)-1, *) += rebin(simul_scan, ind_scans(s+1)-ind_scans(s), nb)

      for b = 0L, nb - 1 do if finite(maxnoise(s, b)) gt 0 then begin
         ind_sort = sort(indxy_scan(*, b))
         ind_uniq = uniq(indxy_scan(ind_sort, b))
         nu = n_elements(ind_uniq)
         ind_mid = ([0, ind_uniq(0:nu-2)+1] + ind_uniq) / 2
         cube_signal(dimxy*b+dimxynb*s+indxy_scan(ind_sort(ind_mid), b)) -= simul_scan(ind_sort(ind_mid))
      endif
   endfor

   if loop lt nloop - 1 then begin
      map_scans = total(cube_signal * cube_weight_ref, 3, /nan)
      map_scans(wdef_scans) /= weight_scans(wdef_scans)
      map_scans(wz_scans) = !values.f_nan

      if loop lt nloop / 2 - 1 then begin
         if n_even ge 2 then begin
            ima_even = total(map_scans(*, *, ind_even) * weight_scans(*, *, ind_even), 3, /nan)
            ima_even(w_even) /= weight_even(w_even)
         endif else ima_even = map_scans(*, *, ind_even)
         if n_odd ge 2 then begin
            ima_odd = total(map_scans(*, *, ind_odd) * weight_scans(*, *, ind_odd), 3, /nan)
            ima_odd(w_odd) /= weight_odd(w_odd)
         endif else ima_odd = map_scans(*, *, ind_odd)
      endif else begin
         map_ref = total(map_scans(*, *, ind_valid) * weight_scans(*, *, ind_valid), 3, /nan)
         map_ref(wdef_ref) /= weight_ref(wdef_ref)
         map_ref(wz_ref) = !values.f_nan
      endelse
   endif
endfor



;;; then destripe by segments of 4 subscans:

nloop = 4
if keyword_set(galactic) then for loop = 0, nloop - 1 do begin
   if loop lt nloop / 2 then nsubs_crit = 9999 else nsubs_crit = 4

   if size_deg gt size_lim_ref then begin
      map_scans = total(cube_signal * cube_weight_ref, 3, /nan)
      map_scans(wdef_scans) /= weight_scans(wdef_scans)
      map_scans(wz_scans) = !values.f_nan
   endif else begin
      ima_signal = total(total(cube_signal(*, *, *, ind_valid) $
         * cube_weight_ref(*, *, *, ind_valid), 4, /nan), 3, /nan)
      ima_signal(wdef_ref) /= weight_ref(wdef_ref)
      ima_signal(wz_ref) = !values.f_nan
      ima_weight = weight_ref
   endelse

   for s = 0, nscans - 1 do begin
      wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans le ind_scans(s+1) - 1, csubs)
      nseg = round(csubs / float(nsubs_crit))
      if nseg gt 1 then begin
         nsubs_cut = max([2 * round(csubs / float(nseg) / 2.), 2])
         nseg = ceil(csubs / float(nsubs_cut))
         ind_start = ind_subscans(wsubs(nsubs_cut*lindgen(nseg)))
         ind_end = [ind_start(1:nseg-1) - 1, ind_scans(s+1) - 1]
         subs_1 = wsubs(nsubs_cut*lindgen(nseg))
         subs_2 = [subs_1(1:nseg-1) - 1, wsubs(csubs-1)]
      endif else begin
         nseg = 1
         ind_start = [ind_scans(s)]
         ind_end = [ind_scans(s+1) - 1]
         subs_1 = [wsubs(0)]
         subs_2 = [wsubs(csubs-1)]
      endelse

      if size_deg gt size_lim_ref then begin
         if (s mod 2) eq 0 then begin
            ima_weight = weight_odd
            if n_odd ge 2 then begin
               ima_signal = total(map_scans(*, *, ind_odd) * weight_scans(*, *, ind_odd), 3, /nan)
               ima_signal(w_odd) /= weight_odd(w_odd)
            endif else ima_signal = map_scans(*, *, ind_odd)
         endif else begin
            ima_weight = weight_even
            if n_even ge 2 then begin
               ima_signal = total(map_scans(*, *, ind_even) * weight_scans(*, *, ind_even), 3, /nan)
               ima_signal(w_even) /= weight_even(w_even)
            endif else ima_signal = map_scans(*, *, ind_even)
         endelse
      endif

      for seg = 0, nseg - 1 do begin
         indxy_scan = indxy_coarse(ind_start(seg):ind_end(seg), *)
         time_scan = float(-time_s(ind_start(seg)) + time_s(ind_start(seg):ind_end(seg)))
         w = where(vscan(ind_start(seg):ind_end(seg)) gt vscan_as_s / 20.)
         indxy_fit = indxy_scan(w, *)
         time_fit = time_scan(w)

         for b = 0L, nb - 1 do if finite(maxnoise(s, b)) gt 0 then begin
            wxy = where(cube_weight_ref(*, *, b, s) gt 0 and cube_subscan(*, *, b, s) ge subs_1(seg) + 1 $
               and cube_subscan(*, *, b, s) le subs_2(seg) + 1 and ima_weight gt 0., cxy)
            if cxy ge 100 then begin
               ima_diff += !values.f_nan
               ima_diff(wxy) = cube_signal(dimxy*b+dimxynb*s+wxy) - ima_signal(wxy)
               fit_scan = absdev_fit(time_fit, ima_diff(indxy_fit(*, b)), coef=coef, noise_level=noise_level, $
                  /sym, /dyn, ns=ns)
               simul_scan = coef(0) + coef(1) * time_scan
               simul(ind_start(seg):ind_end(seg), b) += simul_scan

               if (keyword_set(debug) and b eq bref and seg eq nseg / 2) then begin
                  time_2 = systime(/seconds)
                  t_proc += (time_2 - time_1)
                  wset, 0
                  w = where(finite(ima_diff(indxy_fit(*, b))) gt 0)
                  otherplot, time_fit(w), ima_diff(indxy_fit(w, b)), xtitle='time (s)', $
                     ytitle='signal (Jy/bolometer)', title='bolometer '+chain(bref) $
                     +', scan '+chain(s)+', segment '+chain(seg)+': observed - simulated from map', psym=-7
                  otherplot, /over, color=10, time_scan, simul_scan
                  res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
                  time_1 = systime(/seconds)
               endif

               ind_sort = sort(indxy_scan(*, b))
               ind_uniq = uniq(indxy_scan(ind_sort, b))
               nu = n_elements(ind_uniq)
               ind_mid = ([0, ind_uniq(0:nu-2)+1] + ind_uniq) / 2
               cube_signal(dimxy*b+dimxynb*s+indxy_scan(ind_sort(ind_mid), b)) -= simul_scan(ind_sort(ind_mid))
            endif
         endif

      endfor
   endfor
endfor



;;; then destripe subscan by subscan:

amp_simul = fltarr(nsubscans)
amp_error = fltarr(nsubscans)
error_fit = fltarr(nsubscans, nb)
ind_excl = intarr(nsubscans)
flag_bol = intarr(nsubscans, nb)


nloop = 0
repeat begin
   if (size_deg gt size_lim_ref and nloop lt nloop_switch) then begin
      map_scans = total(cube_signal * cube_weight_ref, 3, /nan)
      map_scans(wdef_scans) /= weight_scans(wdef_scans)
      map_scans(wz_scans) = !values.f_nan
   endif else begin
      ima_signal = total(total(cube_signal(*, *, *, ind_valid) $
         * cube_weight_ref(*, *, *, ind_valid), 4, /nan), 3, /nan)
      ima_signal(wdef_ref) /= weight_ref(wdef_ref)
      ima_signal(wz_ref) = !values.f_nan
      ima_weight = weight_ref
   endelse

   flag_bol_cur = 0
   for s = 0, nscans - 1 do begin
      wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans le ind_scans(s+1) - 1, csubs)
      subs_visu = wsubs(csubs / 2)

      if (size_deg gt size_lim_ref and nloop lt nloop_switch) then begin
         if (s mod 2) eq 0 then begin
            ima_weight = weight_odd
            if n_odd ge 2 then begin
               ima_signal = total(map_scans(*, *, ind_odd) * weight_scans(*, *, ind_odd), 3, /nan)
               ima_signal(w_odd) /= weight_odd(w_odd)
            endif else ima_signal = map_scans(*, *, ind_odd)
         endif else begin
            ima_weight = weight_even
            if n_even ge 2 then begin
               ima_signal = total(map_scans(*, *, ind_even) * weight_scans(*, *, ind_even), 3, /nan)
               ima_signal(w_even) /= weight_even(w_even)
            endif else ima_signal = map_scans(*, *, ind_even)
         endelse
      endif

      for subs = wsubs(0), wsubs(csubs-1) do begin
         nsubs = ind_subscans(subs+1) - ind_subscans(subs)
         indxy_subs = indxy_coarse(ind_subscans(subs):ind_subscans(subs+1)-1, *)
         simul_subs = fltarr(ind_subscans(subs+1)-ind_subscans(subs), nb)
         time_subs = float(-time_s(ind_subscans(subs)) + time_s(ind_subscans(subs):ind_subscans(subs+1)-1))
         wdef_fit = where(vscan(ind_subscans(subs):ind_subscans(subs+1)-1) gt vscan_as_s / 20.)
         indxy_fit = indxy_subs(wdef_fit, *)
         time_fit = time_subs(wdef_fit)


         ;;; break down deviant subscans into several pieces:
         if (ind_excl(subs) gt 0 and nsubs gt 1000L and (not keyword_set(galactic))) then begin
            slice_diff_aver = !values.f_nan + fltarr(nsubs, nb)
            for b = 0L, nb - 1 do if finite(maxnoise(s, b)) gt 0 then begin
               wxy = where(cube_subscan(*, *, b, s) eq subs + 1 and ima_weight gt 0., cxy)
               if cxy ge 10 then begin
                  ima_diff += !values.f_nan
                  ima_diff(wxy) = cube_signal(dimxy*b+dimxynb*s+wxy) - ima_signal(wxy)
                  slice_diff_aver(wdef_fit, b) = ima_diff(indxy_fit(*, b))
               endif
            endif
            weight_diff_aver = total(finite(slice_diff_aver), 2)
            slice_diff_aver = total(slice_diff_aver, 2, /nan) / weight_diff_aver
            w = where(weight_diff_aver lt nb / 2 or flag_nza(ind_subscans(subs):ind_subscans(subs+1)-1) lt 1, cw)
            if cw ge 1 then slice_diff_aver(w) = !values.f_nan

            nt_seg = 500L
            nseg = floor(nsubs / nt_seg)
            fit_subscan = !values.f_nan + fltarr(nsubs)
            for kseg = 0L, nseg - 1L do begin
               it_i = kseg * nt_seg
               it_f = (kseg + 1) * nt_seg - 1
               if kseg eq nseg - 1 then it_f = nsubs - 1
               w = where(finite(slice_diff_aver(it_i:it_f)) gt 0, cw)
               if cw ge 10 then begin
                  if kseg gt 0L then off_i = 50 else off_i = 0
                  if kseg lt nseg - 1L then off_f = 50 else off_f = 0
                  time_seg = time_subs(it_i:it_f)
                  fit_seg = absdev_fit(time_seg, slice_diff_aver(it_i:it_f), coef=coef, noise_level=noise_level, /sym, /dyn, ns=ns)
                  fit_subscan(it_i+off_i:it_f-off_f) = coef(0) + coef(1) * time_seg(off_i:it_f-it_i-off_f)
               endif
            endfor
            wz = where(finite(fit_subscan) lt 1, complement=wdef)
            fit_subscan(wz) = interpol(fit_subscan(wdef), time_subs(wdef), time_subs(wz))
            for b = 0L, nb - 1 do if finite(maxnoise(s, b)) gt 0 then simul_subs(*, b) += fit_subscan
         endif


         for b = 0L, nb - 1 do if finite(maxnoise(s, b)) gt 0 then begin
            wxy = where(cube_subscan(*, *, b, s) eq subs + 1 and ima_weight gt 0., cxy)
            if cxy ge 10 then begin
               ima_diff += !values.f_nan
               ima_diff(wxy) = cube_signal(dimxy*b+dimxynb*s+wxy) - ima_signal(wxy)
               fit_subs = absdev_fit(time_fit, ima_diff(indxy_fit(*, b)) - simul_subs(wdef_fit, b), coef=coef, $
                  noise_level=noise_level, /sym, /dyn, ns=ns)
               werr = where(finite(ima_diff(indxy_fit(*, b))) gt 0)
               error_fit(subs, b) = sqrt(mean((ima_diff(indxy_fit(werr, b)) - fit_subs(werr))^2.))
               simul_subs(*, b) += (coef(0) + coef(1) * time_subs)

               if (keyword_set(debug) and subs eq subs_visu and b eq bref) then begin
                  time_2 = systime(/seconds)
                  t_proc += (time_2 - time_1)
                  wset, 0
                  w = where(finite(ima_diff(indxy_fit(*, b))) gt 0)
                  otherplot, time_fit(w), ima_diff(indxy_fit(w, b)), xtitle='time (s)', $
                     ytitle='signal (Jy/bolometer)', title='bolometer '+chain(bref) $
                     +', subscan '+chain(subs)+': observed - simulated from map', psym=-7
                  otherplot, /over, color=10, time_subs, simul_subs(*, b)
                  res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
                  time_1 = systime(/seconds)
               endif

               ind_sort = sort(indxy_subs(*, b))
               ind_uniq = uniq(indxy_subs(ind_sort, b))
               nu = n_elements(ind_uniq)
               ind_mid = ([0, ind_uniq(0:nu-2)+1] + ind_uniq) / 2
               cube_signal(dimxy*b+dimxynb*s+indxy_subs(ind_sort(ind_mid), b)) -= simul_subs(ind_sort(ind_mid), b)
            endif else begin
               error_fit(subs, b) = 0.
               simul_subs(*, b) = 0.
            endelse
         endif else begin
            error_fit(subs, b) = 0.
            simul_subs(*, b) = 0.
         endelse

         simul(ind_subscans(subs):ind_subscans(subs+1)-1, *) += simul_subs
         wb = where(error_fit(subs, *) gt 0., cb)
         if cb ge 1 then begin
            amp_simul(subs) = median(sqrt(total((simul_subs(*, wb) - (median(simul_subs(*, wb), dimension=1) ## (fltarr(nsubs) + 1.)))^2., 1) $
               / float(nsubs)) / maxnoise(s, wb), /even)
            amp_error(subs) = median(error_fit(subs, wb) / maxnoise(s, wb), /even)
         endif else amp_error(subs) = 0.
         if (cb ge 0.7 * nb and nloop le 1) then begin
            m = median(error_fit(subs, wb), /even)
            dev = sqrt(mean((error_fit(subs, wb) - m)^2.))
            w = where(error_fit(subs, wb) lt m + 3. * dev)
            dev = sqrt(mean((error_fit(subs, wb(w)) - m)^2.))
            w = where(error_fit(subs, *) gt m + 60. * dev and flag_bol(subs, *) lt 1, cw)
            if cw ge 1 then begin
               flag_bol_cur = 1
               flag_bol(subs, w) = 1
               signal(ind_subscans(subs):ind_subscans(subs+1)-1, w) = !values.f_nan
               for ib = 0L, cw - 1L do cube_weight_ref(indxy_subs(*,w(ib))+dimxy*w(ib)+dimxynb*s) = 0.
               if keyword_set(visu) then print, '     subscan ' + chain(subs) + ': excluding bolometer ' $
                  + chain(w) + ' (ndev = ' + chain((error_fit(subs, w) - m) / dev, 1) + ')'
            endif
         endif
      endfor
   endfor


   if (not keyword_set(disjoint)) then if nloop lt max_nloop - 1 then begin
      if nscans lt 4 then thres_dev = 5. else thres_dev = 3.5
      amp_dev = amp_error
      for s = 0, nscans - 1 do begin
         wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1) and amp_error gt 0., csubs)
         if csubs lt 1 then begin
            print, ''
            print, 'not enough resolution elements !'
            print, 'Use the /minimap option if the observations'
            print, 'were taken in the small-map or mini-map mode,'
            print, 'or if the field size is <~ 0.1 degree.'
            print, 'Otherwise, please check the mask settings in HIPE.'
            print, 'Exit.'
            retall
         endif
         med_amp = median(amp_error(wsubs), /even)
         amp_dev(wsubs) -= med_amp
      endfor
      wsubs = where(amp_error gt 0.)
      sig_amp = sqrt(mean(amp_dev(wsubs)^2.))
      ndev = amp_dev / sig_amp
      w = where((abs(ndev) gt thres_dev or amp_error le 0.) and ind_excl lt 1, cw)
      if cw ge 1 then begin
         print, '     excluding deviant subscan ' + chain(w) + ' from reference map: ndev = ' + chain(ndev(w), 1) + crs
         ind_excl(w) = 1
         if keyword_set(debug) then begin
            time_2 = systime(/seconds)
            t_proc += (time_2 - time_1)
            wset, 0
            otherplot, lindgen(nsubscans), amp_dev, psym=4, xtitle='subscan index', ytitle='normalized error'
            otherplot, /over, lstyle=2, color=10, lindgen(nsubscans), fltarr(nsubscans) + 2. * sig_amp
            otherplot, /over, lstyle=2, color=10, lindgen(nsubscans), fltarr(nsubscans) + 3. * sig_amp
            otherplot, /over, lstyle=2, color=10, lindgen(nsubscans), fltarr(nsubscans) + 4. * sig_amp
            if cw ge 1 then otherplot, /over, color=10, w, amp_dev(w), psym=4
            res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
            time_1 = systime(/seconds)
         endif
      endif
      for ks = 0L, cw - 1L do begin
         subs = w(ks)
         wz = where(cube_subscan eq subs + 1, cz)
         if cz ge 1 then cube_weight_ref(wz) = 0.
      endfor
      if (cw ge 1 or keyword_set(flag_bol_cur)) then begin
         weight_scans = total(cube_weight_ref, 3)
         wdef_scans = where(weight_scans gt 0., complement=wz_scans)
         weight_ref = total(weight_scans(*, *, ind_valid), 3)
         if nscans lt 4 then begin
            wz = where(weight_even * weight_odd le 0.)
            weight_ref(wz) = 0.
         endif
         wdef_ref = where(weight_ref gt 0., complement=wz_ref)
      endif
   endif

   nloop += 1
   if keyword_set(debug) then print, mean(amp_error(wsubs)), mean(amp_simul(wsubs))
endrep until ((nloop ge min_nloop and (mean(amp_error(wsubs)) le 1. or mean(amp_simul(wsubs)) le 0.1)) or nloop ge max_nloop)



;;; then update signal and drift:

for s = 0, nscans - 1 do begin
   time_ref = time_s(ind_scans(s))
   simul_scan = simul(ind_scans(s):ind_scans(s+1)-1, *)
   wnza = where(flag_nza(ind_scans(s):ind_scans(s+1)-1) eq 0, cnza, complement=wdef, ncomplement=cdef)
   if cnza ge 1 then begin
      tab_simul = transpose([simul_scan(wdef(0), *), simul_scan(wdef, *), simul_scan(wdef(cdef-1), *)])
      tab_time = float(-time_ref + [time_s(ind_scans(s))-maxtimediff, time_s(ind_scans(s)+wdef), time_s(ind_scans(s+1)-1)+maxtimediff])
      ind_time = interpol(findgen(cdef+2), tab_time, float(-time_ref + time_s(ind_scans(s):ind_scans(s+1)-1)))
      simul_scan = transpose(interpolate(tab_simul, ind_time))
   endif
   signal(ind_scans(s):ind_scans(s+1)-1, *) -= simul_scan
   cscan = ind_scans(s+1) - ind_scans(s)
   tab_simul = transpose([simul_scan(0, *), simul_scan, simul_scan(cscan-1, *)])
   tab_time = float(-time_ref + [time_s(ind_scans(s))-maxtimediff, time_s(ind_scans(s):ind_scans(s+1)-1), time_s(ind_scans(s+1)-1)+maxtimediff])
   ind_time = interpol(findgen(cscan+2), tab_time, -float(time_ref) + time_save(ind_sc_coarse(s):ind_sc_coarse(s+1)-1))
   drifts_all(ind_sc_coarse(s):ind_sc_coarse(s+1)-1, *) += transpose(interpolate(tab_simul, ind_time))
endfor


;;; match brightness of distinct observations with small overlap:
if keyword_set(disjoint) then begin
   save, file=file_signal, signal, /xdr
   tab_match = fltarr(nscans)

   ;;; first create single scan maps:
   dimxy = dimx * dimy
   map_scans = fltarr(dimx, dimy, nscans)
   weight_scans = fltarr(dimx, dimy, nscans)
   error_scans = fltarr(dimx, dimy, nscans)
   for i = 0, nscans - 1 do begin
      prepare_proj_fine, file_scalars=file_scalars, scan_next=i+1, ind_scans=ind_scans
      proj_map, map_signal, map_weight, map_error, file_scalars=file_scalars, maxnoise=maxnoise, $
         ind_scans=ind_scans, /comp_errors
      map_scans(*, *, i) = map_signal
      weight_scans(*, *, i) = map_weight
      error_scans(*, *, i) = map_error
   endfor
   prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans

   ;;; change scan order to have the maximum overlap between successive scans:
   tab_cw = lonarr(nscans, nscans)
   for j = 1, nscans - 1 do for i = 0, j - 1 do begin
      w_over = where(weight_scans(*, *, i) * weight_scans(*, *, j) gt 0., cw)
      tab_cw(i, j) = cw
   endfor
   m = max(tab_cw, ind)
   ind_overlap_scans = [ind mod nscans, ind / nscans]
   tab_cw(ind) = 0
   ncur = 2
   while ncur lt nscans do begin
      tab_cw_check = lonarr(nscans, nscans)
      for k = 0, ncur - 1 do begin
         tab_cw_check(ind_overlap_scans(k), *) = tab_cw(ind_overlap_scans(k), *)
         tab_cw_check(*, ind_overlap_scans(k)) = tab_cw(*, ind_overlap_scans(k))
      endfor
      m = max(tab_cw_check, ind)
      if min(abs(ind_overlap_scans - (ind mod nscans))) eq 0 then ind_cur = ind / nscans else ind_cur = (ind mod nscans)
      ind_overlap_scans = [ind_overlap_scans, ind_cur]
      tab_cw(ind) = 0
      ncur += 1
   endwhile

   ;;; then match brightness of sucessive scans:
   for k = 1, nscans - 1 do begin
      i = ind_overlap_scans(k)
      if k gt 1 then begin
         map_weight_combined = total(weight_scans(*, *, ind_overlap_scans(0:k-1)), 3)
         map_signal_combined = total(map_scans(*, *, ind_overlap_scans(0:k-1)) * weight_scans(*, *, ind_overlap_scans(0:k-1)), 3) $
            / map_weight_combined
         map_error_combined = total(error_scans(*, *, ind_overlap_scans(0:k-1)) * weight_scans(*, *, ind_overlap_scans(0:k-1)), 3) $
            / map_weight_combined
      endif else begin
         map_weight_combined = weight_scans(*, *, ind_overlap_scans(0))
         map_signal_combined = map_scans(*, *, ind_overlap_scans(0))
         map_error_combined = error_scans(*, *, ind_overlap_scans(0))
      endelse
      w_over = where(weight_scans(*, *, i) * map_weight_combined gt 0.)
      m = min([median(weight_scans(i*dimxy+w_over)), median(map_weight_combined(w_over))])
      w_over = where(weight_scans(*, *, i) gt m / 2. and map_weight_combined gt m / 2.)
      m = min([median(weight_scans(i*dimxy+w_over)), median(map_weight_combined(w_over))])
      w_over = where(weight_scans(*, *, i) gt 0.7 * m and map_weight_combined gt 0.7 * m $
         and error_scans(*, *, i) lt 2. * map_error_combined and map_error_combined lt 2. * error_scans(*, *, i))
      tab_match(i) = median(map_scans(i*dimxy+w_over) - map_signal_combined(w_over))
      if keyword_set(visu) then print, 'offset of scan ' + chain(i) + ' wrt preceding scan(s) = ' + chain(tab_match(i), 5)
      map_scans(*, *, i) -= tab_match(i)
   endfor
   for i = 1, nscans - 1 do signal(ind_scans(i):ind_scans(i+1)-1, *) -= tab_match(i)

endif



if (not keyword_set(galactic)) then begin
   w = where(finite(signal) gt 0)
   med = median(signal(w))
   sig = sqrt(mean((signal(w) - med)^2.))
   for k = 0, 9 do begin
      wl = where(signal(w) - med lt 2.5 * sig)
      med = median(signal(w(wl)))
      sig = sqrt(mean((signal(w(wl)) - med)^2.))
   endfor
endif else begin
   nsamp = nb * (2L + ceil(nt / 1000))
   w = where(finite(signal) gt 0)
   ind_sort = sort(signal(w))
   med = median(signal(w(ind_sort(0:nsamp))))
endelse
signal -= med


save, file=file_signal, signal, /xdr

if max(flag_bol) gt 0 then begin
   if (not keyword_set(disjoint)) then prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
   prepare_proj_coarse, file_scalars=file_scalars
endif

end
