function subtract_aver_drift, drifts_all, flag_valid_scans=flag_valid_scans, maxnoise=maxnoise, $
         highfnoise=highfnoise, ind_sc_coarse=ind_sc_coarse, ind_scans=ind_scans, ind_subscans=ind_subscans, $
         ind_aver=ind_aver, file_scalars=file_scalars, title=title, min_nsamp=min_nsamp, detect_gl=detect_gl, $
         t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, parallel=parallel, galactic=galactic, nocross=nocross, iter=iter, $
         size_deg=size_deg, cut_start=cut_start, dr_bolo=dr_bolo, clean=clean, time_1=time_1, t_proc=t_proc


;;; computes average drift, interpolates and optionally subtracts it from the signal

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

;;; copyright 2007-2012 H\'el\`ene Roussel
;;; roussel@iap.fr
;;; Institut d'Astrophysique de Paris - Universit\'e Pierre et Marie Curie, France

;;; download and documentation: http://www2.iap.fr/users/roussel/herschel/

;;; Scanamorphos is distributed under the terms of the GNU General Public License
;;; (http://www.gnu.org/licenses/gpl.html).

;;; July 2010: public release version 1

;;;;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



restore, file_scalars

crs = string(13b) + string(10b)


compute_aver_drift, drifts_save, npairs_save, w_glitches, nb_glitch, maxnoise=maxnoise, highfnoise=highfnoise, $
   ind_scans=ind_scans, ind_aver=ind_aver, file_scalars=file_scalars, min_nsamp=min_nsamp, $
   detect_gl=detect_gl, t_cross_fwhm=t_cross_fwhm, f_fwhm=f_fwhm, parallel=parallel, galactic=galactic, $
   time_1=time_1, t_proc=t_proc


;;; mask glitches and update projection vectors:
if nb_glitch ge 1 then begin
   print, crs + 'removed ' + chain(nb_glitch) + ' sample(s) affected by glitches'
   restore, file_signal
   signal(w_glitches) = !values.f_nan
   save, file=file_signal, signal, /xdr
   signal = 0
   w_glitches = 0

   prepare_proj_coarse, file_scalars=file_scalars

   if (keyword_set(visu) and nb_glitch gt 100) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      restore, file_maps
      if keyword_set(clean) then map_signal = map_signal_uncl
      prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans
      proj_map, map_noglitch, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
      disp_ima, map_signal - map_noglitch, win=4, title='map of residual glitches', $
         weightmap=map_weight, /trunc_max, noise_level=median(maxnoise)
      res = dialog_message('Continue and delete window ?', /center, /information, title='Pause')
      wdelete, 4
      map_noglitch = 0
      map_signal = 0
      map_weigth = 0
      time_1 = systime(/seconds)
   endif
endif



restore, file_times

restore, file_time_coarse


drifts = fltarr(nt)
amp_drifts = fltarr(nscans, nb)

w = where(npairs_save gt 0.)
npairs_min = max([1. / max(maxnoise^2., /nan) / 4., median(npairs_save(w)) / 20])


;;; cancel correction if it is too sparsely determined:

flag_nza_coarse = interpol(flag_nza, float(time_s), time_save)
wnom = where(flag_nza_coarse gt 0., cnom)
wdef = where(flag_nza_coarse gt 0. and npairs_save gt npairs_min, cdef)
ratio_def = float(cdef) / float(cnom)
if ratio_def lt 0.7 then begin
   print, '              average drift too sparsely determined -> correction cancelled.'
   amp_drifts = fltarr(nscans, nb)
   return, amp_drifts
endif


;;; interpolate on fine time grid and suppress potential spikes in average drift (supposed to be smooth):

wz = where(npairs_save le npairs_min, cz)
if cz ge 1 then drifts_save(wz) = !values.f_nan

npairs_interp = fltarr(nt)
for i = 0, nscans - 1 do begin
   time_ref = time_s(ind_scans(i))
   csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
   wsave = ind_sc_coarse(i) + lindgen(csave)
   drifts_sm1 = fltarr(csave)
   drifts_sm2 = fltarr(csave)
   wf = where(finite(drifts_save(wsave)) gt 0)
   drifts_sm1(wf) = median(drifts_save(wsave(wf)), 11)
   drifts_sm2(wf) = median(drifts_save(wsave(wf)), 3)
   wz = where(finite(drifts_save(wsave)) lt 1 or abs(drifts_save(wsave) - drifts_sm1) gt median(maxnoise) $
      or abs(drifts_save(wsave) - drifts_sm2) gt median(maxnoise), cz, complement=wdef, ncomplement=cdef)
   if cz ge 1 then drifts_save(wsave(wz)) = !values.f_nan

   if cdef ge 15 then begin
      if keyword_set(debug) then begin
         wset, 0
         otherplot, time_save(wsave(wf)), drifts_save(wsave(wf)), psym=-4, $
            xtitle='t (s) (coarse grid)', ytitle='(true + excess) drift (Jy/bol.)', $
            title='scan ' + chain(i+1) +  ': original drift in black, interpolated drift in red'
      endif

      if cz ge 1 then begin
         tab_drifts = drifts_save(wsave)
         tab_drifts(wz) = !values.f_nan
         if ratio_def lt 0.95 then wdef = where(smooth(float(finite(tab_drifts)), 15) ge 0.7 and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
         if cdef ge 11 then begin
            tab_drifts(wz) = !values.f_nan
            if cdef ge 3 then begin
               for k = 0, cz - 1 do begin
                  m = min(abs(wz(k) - wdef), ind)
                  flag_end = 0
                  flag_beg = 0
                  if wdef(ind) lt wz(k) then begin
                     ind_1 = ind
                     ind_2 = ind + 1
                     if ind_2 ge cdef then begin
                        ind_1 = ind - 3
                        ind_2 = ind - 2
                        flag_end = 1
                     endif
                  endif else begin
                     ind_1 = ind - 1
                     ind_2 = ind
                     if ind_1 lt 0 then begin
                        ind_1 = ind + 2
                        ind_2 = ind + 3
                        flag_beg = 1
                     endif
                  endelse
                  di = wdef(ind_2) - wdef(ind_1)
                  tabx1 = wdef(ind_1) - reverse(lindgen(min([di+1, wdef(ind_1)+1])))
                  tabx2 = wdef(ind_2) + lindgen(min([di+1, csave-wdef(ind_2)]))
                  taby1 = tab_drifts(tabx1)
                  taby2 = tab_drifts(tabx2)
                  x = [total(tabx1 * finite(taby1)) / total(finite(taby1)), total(tabx2 * finite(taby2)) / total(finite(taby2))]
                  y = [total(taby1, /nan) / total(finite(taby1)), total(taby2, /nan) / total(finite(taby2))]
                  if flag_end eq 1 then y(0) = y(1)
                  if flag_beg eq 1 then y(1) = y(0)
                  tab_drifts(wz(k)) = interpol(y, x, wz(k))
               endfor
               drifts_save(wsave) = tab_drifts
            endif else drifts_save(wsave(wz)) = 0.
            npairs_save(wsave(wz)) = 0.
         endif else begin
            drifts_save(wsave) = 0.
            npairs_save(wsave) = 0.
         endelse
      endif

      if keyword_set(debug) then begin
         otherplot, /over, color=10, psym=-4, time_save(wsave), drifts_save(wsave)
         res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      endif

      tab_time = [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)]
      tab_drifts = [drifts_save(wsave(0)), drifts_save(wsave), drifts_save(wsave(csave-1))]
      subs_drifts = interpol(tab_drifts, tab_time, float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
      drifts(ind_scans(i):ind_scans(i+1)-1) = subs_drifts
      npairs_interp(ind_scans(i):ind_scans(i+1)-1) = interpol(npairs_save(wsave), $
         -float(time_ref) + time_save(wsave), float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
   endif else begin
      drifts_save(wsave) = 0.
      npairs_save(wsave) = 0.
   endelse
endfor

restore, file_vscan
w = where(vscan lt vscan_as_s / 20. or vscan gt 1.5 * vscan_as_s, cw)
if cw ge 1 then drifts(w) = !values.f_nan
vscan = 0




;;; remove drift common to all cross-scans
;;; (solution to the problem of deriving the absolute drift from the drift differences matrix
;;; is not unique, and is affected by oscillations with a periodicity of 2 scans, i.e. 1 cross-scan,
;;; that can be removed in this way):

restore, file_signal
signal_save = signal
restore, file_maps
map_signal_save = map_signal
map_weight_save = map_weight
if (not keyword_set(map_signal_uncl)) then map_signal_uncl = 0
map_uncl_save = map_signal_uncl

restore, file_pixcoords


if (nscans ge 2 and (not keyword_set(nocross))) then begin

   map_drift = fltarr(dimx, dimy, nscans)
   weight_drift = fltarr(dimx, dimy, nscans)
   if size_deg lt 1. then ns_lim = 30 else ns_lim = 15

   signal = rebin(drifts, nt, nb)
   w = where(npairs_interp le npairs_min, cw)
   if cw ge 1 then signal(w, *) = !values.f_nan
   simnoise = 1. + fltarr(nscans, nb)

   ;;; keep only the central part of the array to project the drift,
   ;;; in order to suppress the noise due to coverage non-uniformities and scan leg edges:
   if keyword_set(dr_bolo) then for i = 0, nscans - 1 do begin
      it_mid = (ind_scans(i)+ind_scans(i+1)-1) / 2
      dx = abs(reform(x_bolos(it_mid, *)) - median(x_bolos(it_mid, *), /even))
      dy = abs(reform(y_bolos(it_mid, *)) - median(y_bolos(it_mid, *), /even))
      r = sqrt(dx^2. + dy^2.)
      w = where(r gt dr_bolo)
      signal(ind_scans(i):ind_scans(i+1)-1, w) = !values.f_nan
      simnoise(i, w) = !values.f_nan
      w = where(r le dr_bolo)
      simnoise(i, w) *= (1. + r(w) / dr_bolo)
   endfor

   save, file=file_signal, signal, /xdr

   if nscans ge 4 then cut_pcal = cut_start else cut_pcal = 0
   for i = 0, nscans - 1 do begin
      prepare_proj_fine, file_scalars=file_scalars, scan_next=i+1, ind_scans=ind_scans, $
         ind_subscans=ind_subscans, cut_pcal=cut_pcal
      proj_map, file_scalars=file_scalars, maxnoise=simnoise, ind_scans=ind_scans
      restore, file_maps
      map_drift(0, 0, i) = map_signal
      weight_drift(0, 0, i) = map_weight
   endfor
   w = where(weight_drift gt 0.)
   med_weight = median(weight_drift(w))
   w = where(weight_drift lt 0.5 * med_weight)
   map_drift(w) = !values.f_nan
   weight_drift(w) = 0.


   if inst eq 'pacs' then begin
      w_sm = 3
      r_crit = 12
   endif else begin
      w_sm = 5
      r_crit = 25
   endelse

   if nscans lt ns_lim then begin
      if sqrt(dimx * dimy) gt 300 then width = 51L else width = 31L
      tab_width = [width, width, 1L]
      map_mask = map_weight_save
      w = where(map_weight_save gt 0.)
      med_weight = median(map_weight_save(w))
      w = where(map_mask lt 0.5 * med_weight and total(weight_drift, 3) le 0)
      map_mask(w) = !values.f_nan
      wall = where(map_weight_save gt 0.)
      length_all = float(1 + max(wall mod dimx) - min(wall mod dimx))
      length_all *= floor(dimx / length_all)
      for i = 0, nscans - 1 do begin
         wi = where(weight_drift(*, *, i) gt 0.)
         length = 1 + max(wi mod dimx) - min(wi mod dimx)
         if length lt 0.75 * length_all then length = round(dimx / length_all * length) / 4 else length = dimx / 4
         map_drift(*, *, i) = interpol_map(map_drift(*, *, i), dimx=dimx, dimy=dimy, $
            r_crit=r_crit, map_mask=map_mask, length=length)
      endfor

      map_sq = map_drift^2.
      ;;; do NOT use the /edge_truncate keyword in smooth
      map_smooth = smooth(map_drift, tab_width, /nan)
      map_smoothsq = smooth(map_sq, tab_width, /nan)
      map_dev2 = map_smoothsq - map_smooth^2.
      w = where(map_dev2 le 0. or finite(map_drift) lt 1, cw)
      if cw ge 1 then map_dev2(w) = !values.f_nan

      map_ref = total(map_drift / map_dev2, 3, /nan) / total(1. / map_dev2, 3, /nan)
      if nscans eq 3 then w = where(total(weight_drift gt 0., 3) le 2) else w = where(total(weight_drift gt 0., 3) le 1)
      map_ref(w) = !values.f_nan
      map_ref = smooth(map_ref, w_sm, /nan)
      map_ref(w) = !values.f_nan
      w = where(weight_drift le 0.)
      map_drift(w) = !values.f_nan
   endif


   map_excess = median(map_drift, dimension=3, /even)
   weight_excess = total(weight_drift, 3)
   wz_excess = where(finite(map_excess) lt 1)
   map_excess = smooth(map_excess, w_sm, /nan)
   map_excess(wz_excess) = !values.f_nan
   if nscans lt ns_lim then begin
      map_ref = interpol_map(map_ref, dimx=dimx, dimy=dimy, r_crit=r_crit, wmask=wz_excess, /force_x)
      w = where(finite(map_ref) gt 0)
      map_excess(w) = map_ref(w)
   endif


   if keyword_set(debug) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      w = where(finite(map_drift) gt 0, cw)
      ind_sort = sort(map_drift(w))
      min_d = map_drift(w(ind_sort(100*nscans)))
      max_d = map_drift(w(ind_sort(cw-1-100*nscans)))
      if max_d - min_d gt 10. * median(maxnoise) then pow_int=0.5 else pow_int = 1.
      disp_ima, win=10, map_drift(*, *, 0), weightmap=weight_drift(*, *, 0), $
         title='mapping of scan 0 average drift', min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      disp_ima, win=11, map_drift(*, *, 1), weightmap=weight_drift(*, *, 1), $
         title='mapping of scan 1 average drift', min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      if nscans lt ns_lim then begin
         weight_ref = fltarr(dimx, dimy)
         w = where(finite(map_ref) gt 0)
         weight_ref(w) = 10.
         disp_ima, win=9, map_ref, weightmap=weight_ref, title='reference drift map', $
            min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      endif else disp_ima, win=9, map_excess, weightmap=weight_excess, title='excess drift map', $
            min_map=min_d, max_map=max_d, noise_level=0.01, pow_int=pow_int
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      wdelete, 9, 10, 11
      time_1 = systime(/seconds)
   endif


   indxy_bref = floor(interpol(median(x_bolos, dimension=2, /even), float(time_s), time_save)) $
      + dimx * floor(interpol(median(y_bolos, dimension=2, /even), float(time_s), time_save))
   x_bolos = 0
   y_bolos = 0
   simul_excess = map_excess(indxy_bref)


   ;;; PACS data affected by much stronger flicker noise, making the average drift noisier
   ;;; -> smooth it by mapping it like the excess drift:
   if inst eq 'pacs' then begin
      drifts_save_init = fltarr(nt_save)
      dimxy = dimx * dimy
      for i = 0, nscans - 1 do begin
         if keyword_set(cut_start) then if cut_start(i) gt 0 then begin
            prepare_proj_fine, file_scalars=file_scalars, scan_next=i+1, ind_scans=ind_scans
            proj_map, file_scalars=file_scalars, maxnoise=simnoise, ind_scans=ind_scans
            restore, file_maps
            w = where(weight_drift(*, *, i) le 0.)
            map_signal(w) = !values.f_nan
            map_drift(0, 0, i) = map_signal
         endif
         drifts_save_init(ind_sc_coarse(i)) = map_drift(dimxy*i+indxy_bref(ind_sc_coarse(i):ind_sc_coarse(i+1)-1))
      endfor
   endif else drifts_save_init = drifts_save
   drifts_save = drifts_save_init - simul_excess

   if keyword_set(debug) then begin
      wset, 0
      w = where(npairs_save gt npairs_min and finite(drifts_save) gt 0.)
      otherplot, time_save(w), drifts_save_init(w), xtitle='t (s) (coarse grid)', $
         ytitle='average drift (Jy/bol.)', title='(true + excess) drift in black, excess drift in yellow'
      otherplot, /over, color=200, time_save(w), simul_excess(w)
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

      otherplot, time_save(w), drifts_save(w), xtitle='t (s) (coarse grid)', $
         ytitle='average drift (Jy/bol.)', title='true drift (after subtraction of excess component)'
      res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   endif


   for i = 0L, nscans - 1 do begin
      time_ref = time_s(ind_scans(i))
      csave = ind_sc_coarse(i+1) - ind_sc_coarse(i)
      wsave = ind_sc_coarse(i) + lindgen(csave)
      wz = where(npairs_save(wsave) le npairs_min or finite(drifts_save(wsave)) lt 1, cz, $
         complement=wdef, ncomplement=cdef)
      if cdef ge 15 then begin
         if cz ge 1 then begin
            tab_drifts = drifts_save(wsave)
            if ratio_def lt 0.95 then wdef = where(npairs_save(wsave) gt npairs_min and smooth(float(finite(tab_drifts)), 15) ge 0.7 and finite(tab_drifts) gt 0, cdef, complement=wz, ncomplement=cz)
            if cdef ge 3 then begin
               tab_drifts(wz) = !values.f_nan
               for k = 0, cz - 1 do begin
                  m = min(abs(wz(k) - wdef), ind)
                  flag_end = 0
                  flag_beg = 0
                  if wdef(ind) lt wz(k) then begin
                     ind_1 = ind
                     ind_2 = ind + 1
                     if ind_2 ge cdef then begin
                        ind_1 = ind - 3
                        ind_2 = ind - 2
                        flag_end = 1
                     endif
                  endif else begin
                     ind_1 = ind - 1
                     ind_2 = ind
                     if ind_1 lt 0 then begin
                        ind_1 = ind + 2
                        ind_2 = ind + 3
                        flag_beg = 1
                     endif
                  endelse
                  di = wdef(ind_2) - wdef(ind_1)
                  tabx1 = wdef(ind_1) - reverse(lindgen(min([di+1, wdef(ind_1)+1])))
                  tabx2 = wdef(ind_2) + lindgen(min([di+1, csave-wdef(ind_2)]))
                  taby1 = tab_drifts(tabx1)
                  taby2 = tab_drifts(tabx2)
                  x = [total(tabx1 * finite(taby1)) / total(finite(taby1)), total(tabx2 * finite(taby2)) / total(finite(taby2))]
                  y = [total(taby1, /nan) / total(finite(taby1)), total(taby2, /nan) / total(finite(taby2))]
                  if flag_end eq 1 then y(0) = y(1)
                  if flag_beg eq 1 then y(1) = y(0)
                  tab_drifts(wz(k)) = interpol(y, x, wz(k))
               endfor
               drifts_save(wsave) = tab_drifts
            endif else drifts_save(wsave) = 0.
         endif
         tab_time = [-float(timestep), -float(time_ref) + time_save(wsave), float(time_s(ind_scans(i+1)-1) + timestep - time_ref)]
         tab_drifts = [drifts_save(wsave(0)), drifts_save(wsave), drifts_save(wsave(csave-1))]
         drifts(ind_scans(i)) = interpol(tab_drifts, tab_time, float(-time_ref + time_s(ind_scans(i):ind_scans(i+1)-1)))
      endif else begin
         drifts_save(wsave) = 0.
         drifts(ind_scans(i):ind_scans(i+1)-1) = 0.
      endelse
   endfor


endif else begin

   map_drift = fltarr(dimx, dimy, nsubscans)
   weight_drift = fltarr(dimx, dimy, nsubscans)
   signal = rebin(drifts, nt, nb)
   save, file=file_signal, signal, /xdr
   for i = 0L, nsubscans - 1L do begin
      prepare_proj_fine, file_scalars=file_scalars, subscan_next=i+1, ind_scans=ind_scans, ind_subscans=ind_subscans
      proj_map, file_scalars=file_scalars, maxnoise=median(maxnoise)+fltarr(nscans, nb), ind_scans=ind_scans
      restore, file_maps
      map_drift(*, *, i) = map_signal
      weight_drift(*, *, i) = map_weight
   endfor

   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      disp_ima, win=10, map_drift(*, *, 0), title='mapping of subscan 0 average drift'
      disp_ima, win=11, map_drift(*, *, 1), title='mapping of subscan 1 average drift'
      res = dialog_message('   Continue ?   ', /center, /question)
      time_1 = systime(/seconds)
   endif

   dimxy = dimx * dimy
   for s = 0, nscans - 1 do begin
      wsubs = where(ind_subscans ge ind_scans(s) and ind_subscans lt ind_scans(s+1), csubs)
      simul_aver_2 = fltarr(ind_subscans(wsubs(1))-ind_subscans(wsubs(0)))
      for i = 0L, csubs - 2L do begin
         map_aver = fltarr(dimx, dimy) + !values.f_nan
         w = where(weight_drift(*, *, wsubs(i)) gt 0. and weight_drift(*, *, wsubs(i+1)) gt 0.)
         map_aver(w) = (map_drift(w+wsubs(i)*dimxy) + map_drift(w+wsubs(i+1)*dimxy)) / 2.

         simul_aver_1 = map_aver(floor(x_bolos(ind_subscans(wsubs(i)):ind_subscans(wsubs(i)+1)-1, *)) $
            + dimx * floor(y_bolos(ind_subscans(wsubs(i)):ind_subscans(wsubs(i)+1)-1, *)))
         simul_aver_1 = median(simul_aver_1, dimension=2)
         wz = where(finite(simul_aver_1) lt 1, cwz, complement=wdef, ncomplement=cdef)
         if cwz ge 1 then begin
            tab_time = [time_s(ind_subscans(wsubs(i)))-timestep, time_s(ind_subscans(wsubs(i))+wdef), $
               time_s(ind_subscans(wsubs(i)+1)-1)+timestep]
            tab_drift = [simul_aver_1(wdef(0)), simul_aver_1(wdef), simul_aver_1(wdef(cdef-1))]
            simul_aver_1(wz) = interpol(tab_drift, tab_time, time_s(ind_subscans(wsubs(i))+wz))
         endif
         simul_aver_1 -= simul_aver_2
         wsave = where(time_save ge time_s(ind_subscans(wsubs(i))) $
            and time_save le time_s(ind_subscans(wsubs(i)+1)-1))
         drifts_save(wsave) -= interpol(simul_aver_1, $
            time_s(ind_subscans(wsubs(i)):ind_subscans(wsubs(i)+1)-1), time_save(wsave))

         simul_aver_2 = map_aver(floor(x_bolos(ind_subscans(wsubs(i)+1):ind_subscans(wsubs(i)+2)-1, *)) $
            + dimx * floor(y_bolos(ind_subscans(wsubs(i)+1):ind_subscans(wsubs(i)+2)-1, *)))
         simul_aver_2 = median(simul_aver_2, dimension=2)
         wz = where(finite(simul_aver_2) lt 1, cwz, complement=wdef, ncomplement=cdef)
         if cwz ge 1 then begin
            tab_time = [time_s(ind_subscans(wsubs(i)+1))-timestep, time_s(ind_subscans(wsubs(i)+1)+wdef), $
               time_s(ind_subscans(wsubs(i)+2)-1)+timestep]
            tab_drift = [simul_aver_2(wdef(0)), simul_aver_2(wdef), simul_aver_2(wdef(cdef-1))]
            simul_aver_2(wz) = interpol(tab_drift, tab_time, time_s(ind_subscans(wsubs(i+1))+wz))
         endif
         wsave = where(time_save ge time_s(ind_subscans(wsubs(i)+1)) $
            and time_save le time_s(ind_subscans(wsubs(i)+2)-1))
         drifts_save(wsave) -= interpol(simul_aver_2, $
            time_s(ind_subscans(wsubs(i)+1):ind_subscans(wsubs(i)+2)-1), time_save(wsave))
      endfor

      csave = ind_sc_coarse(s+1) - ind_sc_coarse(s)
      wsave = ind_sc_coarse(s) + lindgen(csave)
      wz = where(npairs_save(wsave) le npairs_min or finite(drifts_save(wsave)) lt 1, cz, $
         complement=wdef, ncomplement=cdef)
      if cdef ge 10 then begin
         wdef = wsave(wdef)
         if cz ge 1 then begin
            wz = wsave(wz)
            tab_time = [float(time_s(ind_scans(s)) - timestep), time_save(wdef), float(time_s(ind_scans(s+1)-1) + timestep)]
            tab_drifts = [drifts_save(wdef(0)), drifts_save(wdef), drifts_save(wdef(cdef-1))]
            drifts_save(wz) = interpol(tab_drifts, tab_time, time_save(wz))
         endif
         tab_time = [float(time_s(ind_scans(s)) - timestep), time_save(wsave), float(time_s(ind_scans(s+1)-1) + timestep)]
         tab_drifts = [drifts_save(wsave(0)), drifts_save(wsave), drifts_save(wsave(csave-1))]
         drifts(ind_scans(s):ind_scans(s+1)-1) = interpol(tab_drifts, tab_time, time_s(ind_scans(s):ind_scans(s+1)-1))
      endif else begin
         drifts_save(wsave) = 0.
         drifts(ind_scans(s):ind_scans(s+1)-1) = 0.
      endelse

   endfor
   x_bolos = 0
   y_bolos = 0
endelse


;;; remove zero level of drifts:
off = median(drifts)
drifts -= off
drifts_save -= off


signal = signal_save
save, file=file_signal, signal, /xdr
map_signal = map_signal_save
map_weight = map_weight_save
map_signal_uncl = map_uncl_save
save, file=file_maps, map_signal, map_weight, map_error, map_lowfnoise, map_signal_uncl, /xdr
prepare_proj_fine, file_scalars=file_scalars, ind_scans=ind_scans


w = where(npairs_save gt npairs_min)
med = median(drifts_save(w))
for i = 0, nscans - 1 do begin
   w = where(flag_nza(ind_scans(i):ind_scans(i+1)-1) gt 0)
   amp_drifts(i, *) = 3. * sqrt(mean((drifts(ind_scans(i)+w) - med)^2.))
endfor



;;; visualize result:

if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   window, 0, xs=1200
   w = where(finite(signal) lt 1, cw)
   mask_bolo = 1 + intarr(nt, nb)
   if cw ge 1 then mask_bolo(w) = 0
   signal_aver = total(signal * mask_bolo, 2, /nan)
   norm = total(mask_bolo, 2)
   mask_bolo = 0
   wdef = where(norm gt nb / 5, complement=wz, ncomplement=nz)
   signal_aver(wdef) /= norm(wdef)
   if nz ge 1 then signal_aver(wz) = !values.f_nan
   med = median(signal_aver(wdef))
   dev = sqrt(mean((signal_aver(wdef) - med)^2.))
   for k = 0, 2 do begin
      wdev = where(norm gt nb / 5. and abs(signal_aver - med) lt 6. * dev)
      med = median(signal_aver(wdev))
      dev = sqrt(mean((signal_aver(wdev) - med)^2.))
   endfor
   dev = max([dev, median(maxnoise)])
   norm = 0
   yrange = [-4.*dev, 4.*dev]
   otherplot, time_s(wdef), signal_aver(wdef) - med, charsize=1.5, yrange=yrange, $
      title=title, xtitle='time (s)', ytitle='brightness + offset'
   w = where(npairs_save gt 0)
   otherplot, /over, time_save(w), drifts_save(w) - 2. * dev, psym=4, color=10
   otherplot, /over, time_s, drifts - 2. * dev, color=100
   for i = 0, nscans - 2 do begin
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1)-1)+fltarr(2), yrange
      otherplot, /over, color=200, lstyle=5., time_s(ind_scans(i+1))+fltarr(2), yrange
   endfor
   if nsubscans le 30 then begin
      otherplot, /over, time_s, dev * (13.8 * flag_nza - 10.), color=200
      for subs = 0L, nsubscans - 1 do otherplot, /over, time_s(ind_subscans(subs))+[0.,0.], yrange, color=30, lstyle=2.
   endif
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')

   yrange = [-4.*dev, 10.*dev]
   otherplot, time_s(wdef), signal_aver(wdef) - med, charsize=1.5, yrange=yrange, $
      title='zoom: raw drift in grey, excess drift in yellow, corrected drift below', $
      xtitle='time (s)', ytitle='brightness + offset', $
      xrange=time_s([ind_subscans(0), ind_subscans(min([4, nsubscans]))-1])
   w = where(npairs_save gt npairs_min)
   otherplot, /over, time_save(w), drifts_save(w) - 2. * dev, psym=4, color=10
   otherplot, /over, time_s, drifts - 2. * dev, color=100
   if (nscans ge 2 and (not keyword_set(nocross))) then begin
      otherplot, /over, time_save(w), drifts_save_init(w) + 5. * dev, psym=4, color=150
      otherplot, /over, time_save(w), drifts_save_init(w) - drifts_save(w) + 5. * dev, psym=4, color=200
   endif
   res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
   time_1 = systime(/seconds)
endif



print, '              drifts amplitude / noise (median, min and max) = ' $
   + chain(median(amp_drifts / maxnoise), 2) + ' ; ' $
   + chain(min(amp_drifts / maxnoise, /nan), 2) + ' to ' + chain(max(amp_drifts / maxnoise, /nan), 2) + crs



;;; update signal and total drifts:

if keyword_set(visu) then begin
   time_2 = systime(/seconds)
   t_proc += (time_2 - time_1)
   res = dialog_message('Subtract average drift ?', /center, /question)
   time_1 = systime(/seconds)
endif else res = 'Yes'

if res eq 'Yes' then begin
   signal -= rebin(drifts, nt, nb)
   drifts_all += rebin(drifts_save, nt_save, nb)
   save, file=file_signal, signal, /xdr

   ;;; visualization of drift:
   if keyword_set(visu) then begin
      time_2 = systime(/seconds)
      t_proc += (time_2 - time_1)
      restore, file_maps
      if keyword_set(clean) then map_signal = map_signal_uncl
      proj_map, map_nodrifts, file_scalars=file_scalars, maxnoise=maxnoise, ind_scans=ind_scans
      title = 'map of iteration ' + chain(iter) + ' drifts'
      if keyword_set(detect_gl) then title = title + ' (+ glitches)'
      map_drifts = map_signal - map_nodrifts
      disp_ima, map_drifts, win=3, title=title, weightmap=map_weight, noise_level=median(maxnoise)
      map_ratio = fltarr(dimx, dimy)
      w = where(map_signal gt median(maxnoise))
      map_ratio(w) = abs(map_drifts(w)) / map_signal(w)
      disp_ima, map_ratio, win=4, title='ratio map of iteration ' + chain(iter) + ' |drifts| / signal', $
         weightmap=map_weight, noise_level=median(maxnoise), max_map=min([max(map_ratio), 0.15])
      if keyword_set(debug) then res = dialog_message('   Continue ?   ', /center, /information, title='Pause')
      time_1 = systime(/seconds)
   endif
endif else amp_drifts *= 0.


return, amp_drifts

end
